﻿using System;
using System.Collections.Generic;
using System.Text;

using JoyHack;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;

namespace JHDrawLib
{
    public class DrawJoyHack
    {
        private List<Phrase> lyr;
        private int _width, _height;
        private int _timeCodeFix; //frames - set from datafile
        private int _timeUserFix; //frames - set from App

        private drawCache[] cache = new drawCache[4];

        public bool _debug;
        public bool disable_cache;

        public bool debug
        {
            get { return this._debug; }
            set {
                this.clear_cache();
                this._debug = value;
            }
        }

        public int width
        {
            get { return this._width; }
            set
            {
                if (value <= 0) { throw new ArgumentOutOfRangeException(); }
                this.clear_cache();
                this._width = value;
            }
        }

        public int height
        {
            get { return this._height; }
            set
            {
                if (value <= 0) { throw new ArgumentOutOfRangeException(); }
                this.clear_cache();
                this._height = value;
            }
        }

        public int timeCodeFix
        {
            get { return this._timeCodeFix; }
            private set { this._timeCodeFix = value; }
        }

        public int timeUserFix
        {
            get { return this._timeUserFix; }
            set { this._timeUserFix = value; }
        }

        public DrawJoyHack(int width, int height)
        {
            this.lyr = new List<Phrase>();
            this.width = width;
            this.height = height;
            this.timeCodeFix = 0;
            this.timeUserFix = 0;
            this.debug = false;
            this.disable_cache = false;
        }

        public void clear_cache ()
        {
            this.cache = new drawCache[4];
        }

        public void load_data(string file)
        {
            this.load_data(file, Encoding.GetEncoding(932));
        }

        public void load_data ( string file , Encoding encode )
        {

            lyr = new List<Phrase>();
            this.clear_cache();

            FileStream fs = new FileStream(file, FileMode.Open);
            StreamReader sr = new StreamReader(fs, encode);

            int showline = 0; //index 0... (all_showline-1)
            int all_showline = 2; // 1...4

            string line;
            WordCond cond = new WordCond();

            cond.color = new ShowColor(
                new Color[8] {Color.White, Color.Black, Color.Red, Color.White,
                                        Color.White, Color.Black, Color.Red, Color.White}
                                        );
            cond.mainfont = new WordFont(
                new Font ("MS UI Gothic", 30, FontStyle.Regular, GraphicsUnit.Pixel), StringAlignment.Near, 1.0D);
            cond.rubyfont = new WordFont(
                new Font("MS UI Gothic", 15, FontStyle.Regular, GraphicsUnit.Pixel), StringAlignment.Near, 1.0D);
            
            Alignment align = Alignment.Default;


            float height = -1;
            int linenum = 0;

            while ((line = sr.ReadLine()) != null) {
                linenum++;

                try
                {

                    line = line.Replace("\r", "").Replace("\n", "");

                    if (line.Contains("<timeCodeFix>"))
                    {
                        string timecodestr = line.Substring("<timeCodeFix>".Length);
                        if (timecodestr.Length == 8)
                        {
                            this.timeCodeFix = new Timecode(timecodestr).allFrames;
                        }
                        else if (timecodestr.Length == 9)
                        {
                            if (timecodestr.Substring(0, 1) == "-")
                            {
                                this.timeCodeFix = -(new Timecode(timecodestr.Substring(1, 8)).allFrames);
                            }
                            else
                            {
                                this.timeCodeFix = +(new Timecode(timecodestr.Substring(1, 8)).allFrames);
                            }
                        }
                        else
                        {
                            throw new Exception("不正な<timeCodeFix>です");
                        }
                        continue;
                    }

                    if (line.Contains("<x_align>"))
                    {
                        string alignstr = line.Substring("<x_align>".Length);
                        byte align_i = byte.Parse(alignstr);
                        if (align_i < 0 || 3 < align_i) { throw new Exception("不正な<x_align>です"); }
                        align = (Alignment)align_i;
                        continue;
                    }

                    if (line.Contains("<time>"))
                    {
                        Phrase p;
                        p = new Phrase(line, showline, all_showline, ref cond, align);
                        if (p.clear)
                        {
                            all_showline = p.all_showline;
                            showline = 0;
                            height = -1;
                        }
                        else
                        {
                            if (height < 0) { height = p.GetMaxHeight(); }
                            p.height = height;
                            showline = (++showline) % all_showline;
                        }
                        lyr.Add(p);
                    }
                }
                catch (Exception e)
                {
                    fs.Close();
                    throw new Exception(linenum.ToString() + " 行目: " + e.Message);
                }
            }

            fs.Close();
        }

        public void drawENT(Graphics g, long millisecond)
        {
            Timecode t = Timecode.FromMillisecond(millisecond);
            this.drawENT(g, t);
        }

        public void drawENT(Graphics g, Timecode time)
        {
            Timecode t = new Timecode( System.Math.Max(0, time.allFrames - this.timeCodeFix - this.timeUserFix));

            if (this.lyr.Count == 0) { return; }

            int i;
            /*
            for (i = 0; i < this.lyr.Count; i++)
            {
                if (this.lyr[i].getStarttime() >= t) { break; }
            }
            i = i - 1; //Now Playing or Last Played
            */
            for (i = this.lyr.Count; i >0; i--)
            {
                if (this.lyr[i-1].getStarttime() < t) { break; }
            }
            i = i - 1; //Now Playing or Last Played

            if ( this.debug )
                g.DrawString(i.ToString() + " " + t.ToString() , new Font("MS UI Gothic", 20), Brushes.Blue, new Point(10, 10));

            int l_showlines;
            Phrase[] drawTemp = { null, null, null, null };
            bool playing_i = false;

            if (i >= 0)
            {
                if (this.lyr[i].getEndtime() >= t)
                { // Now Playing
                    playing_i = true;
                    if (this.lyr[i].clear) { return; }
                }
                l_showlines = this.lyr[i].all_showline;

                if (l_showlines >= 2 || playing_i)
                {
                    if (!this.lyr[i].clear) {
                        drawTemp[this.lyr[i].showline] = this.lyr[i];
                    }
                }

                if (l_showlines >= 3 && !this.lyr[i].clear)
                {
                    if (i - 1 >= 0)
                    {
                        if (!this.lyr[i - 1].clear)
                        {
                            drawTemp[this.lyr[i-1].showline] = this.lyr[i-1];
                            if (i - 2 >= 0 && l_showlines >= 4)
                            {
                                if (!this.lyr[i - 2].clear)
                                {
                                    drawTemp[this.lyr[i-2].showline] = this.lyr[i-2];
                                }
                            }
                        }
                    }
                }
            }
            else
            {
                //if (i + 1 >= this.lyr.Count) { throw new Exception("空の歌詞データです"); }
                l_showlines = this.lyr[i + 1].all_showline;
            }

            if (i + 1 < this.lyr.Count)
            {
                if (!playing_i || l_showlines >= 2)
                {
                    if (!this.lyr[i + 1].clear) { drawTemp[this.lyr[i+1].showline] = this.lyr[i+1]; }
                }
            }

            for (int cnt = 0; cnt < l_showlines; cnt++)
            {
                if (drawTemp[cnt] != null)
                {
                    this.drawPhrase(g, drawTemp[cnt], t,
                        (int)(this.height - drawTemp[cnt].height * 1.8 * (drawTemp[cnt].all_showline - cnt) - 10));
                }
            }
            /*if (i == 6) { i = i; }
            int y = -1;
            int blankheight = -1;
            for (int cnt = 0; cnt < l_showlines; cnt++)
            {
                if (drawTemp[cnt] != null)
                {
                    if (y == -1)
                    {
                        y = (int)(this.height - drawTemp[cnt].GetMaxHeight() * 1.8 * (drawTemp[cnt].all_showline - cnt) -10);
                    }
                    this.drawPhrase(g, drawTemp[cnt], t, y);
                    y += (int)(drawTemp[cnt].GetMaxHeight() * 1.8);
                    blankheight = (int)(drawTemp[cnt].GetMaxHeight() * 1.8);
                }
                else
                {
                    if (blankheight != -1 && y != -1)
                    {
                        y += blankheight;
                    }
                }
            }
             * */
        }

        private void drawPhrase(Graphics g_moto, Phrase phr, Timecode time, int y)
        {

            if (phr.clear) { return; }

            if (this.cache[phr.showline].phr != phr || this.disable_cache) //not hit...
            {
                this.cache[phr.showline] = new drawCache(phr);

                //moto
                Bitmap bmp1 = new Bitmap(this.width, this.height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
                Graphics g = Graphics.FromImage(bmp1);
                this.cache[phr.showline].bmp_moto = bmp1;

                //effected
                Bitmap bmp2 = new Bitmap(this.width, this.height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
                Graphics g2 = Graphics.FromImage(bmp2);
                this.cache[phr.showline].bmp_effe = bmp2;

                g.SmoothingMode = SmoothingMode.AntiAlias;
                g.PixelOffsetMode = PixelOffsetMode.HighQuality;

                g2.SmoothingMode = SmoothingMode.AntiAlias;
                g2.PixelOffsetMode = PixelOffsetMode.HighQuality;

                //int y = this.height - (int)(fnt.Height *1.8) * (phr.all_showline-phr.showline);

                // C#: DrawStringとAddStringでの文字描画サイズの違い
                // http://dobon.net/vb/bbs/log3-10/5479.html

                StringFormat allsf = new StringFormat();
                List<Rectangle> width_arr = new List<Rectangle>();

                allsf.FormatFlags = StringFormatFlags.NoClip | StringFormatFlags.NoWrap;
                allsf.Trimming = StringTrimming.None;

                int all_width = 0;
                for (int i = 0; i < phr.words.Count; i++)
                {
                    if (phr.words[i].str.Length == 0)
                    {
                        width_arr.Add(new Rectangle(0, 0, 0, 0));
                        all_width += 0;
                        continue;
                    }

                    allsf.SetMeasurableCharacterRanges(
                        new CharacterRange[] { new CharacterRange(0, phr.words[i].str.Length) });

                    Region[] rrr = g.MeasureCharacterRanges
                        (phr.words[i].str.Replace(' ', '_').Replace('　' /*full-width space*/, '駄'),
                        phr.words[i].cond.mainfont.font, new RectangleF(0, 0, this.width, this.height), allsf);
                    Rectangle rrrect = Rectangle.Truncate(rrr[0].GetBounds(g));

                    rrrect.Width = (int)(rrrect.Width * phr.words[i].cond.mainfont.adjust_width);
                    rrrect.Width += 2;

                    if (phr.words[i].cond.width >= 0)
                    {
                        rrrect.Width = phr.words[i].cond.width;
                    }

                    all_width += rrrect.Width;
                    width_arr.Add(rrrect);
                }

                int x = -1;
                Alignment align = phr.align;
                if (align == Alignment.Default)
                {
                    if ((phr.all_showline % 2) == 0)
                    {
                        if ((phr.showline % 2) == 0) { align = Alignment.Left; }
                        else
                        {
                            align = Alignment.Right;
                        }
                    }
                    else if (phr.all_showline == 3)
                    {
                        if (phr.showline == 0) { align = Alignment.Left; }
                        else if (phr.showline == 1)
                        {
                            align = Alignment.Center;
                        }
                        else if (phr.showline == 2)
                        {
                            align = Alignment.Right;
                        }
                    }
                    else if (phr.all_showline == 1)
                    {
                        align = Alignment.Left;
                    }
                }

                switch (align)
                {
                    case Alignment.Left:
                        x = 20;
                        break;
                    case Alignment.Center:
                        x = Math.Max((this.width - all_width) / 2, 20);
                        break;
                    case Alignment.Right:
                        x = Math.Max(this.width - 20 - all_width, 20);
                        break;
                    default:
                        throw new Exception("表示行がなんかおかしいです");
                        break;
                }

                //int afterx = 0;

                //foreach (Word w in phr.words)
                //{
                for (int i = 0; i < phr.words.Count; i++)
                {
                    Word w = phr.words[i];

                    int lsi;
                    int newX = x;
                    int oldX = x;

                    if (w.str.Length == 0)
                    {
                        //goto skip_drawing;
                        lsi = 0;
                    }
                    else
                    {
                        Font fnt = w.cond.mainfont.font;
                        StringFormat sf = new StringFormat();

                        sf.FormatFlags = StringFormatFlags.NoClip | StringFormatFlags.NoWrap;
                        sf.Trimming = StringTrimming.None;
                        sf.LineAlignment = w.cond.mainfont.line_align;

                        Rectangle strrect = width_arr[i];
                        lsi = strrect.Width;
                        oldX = x;
                        x -= strrect.X;

                        if (strrect.Height <= 0) { goto skip_drawing; }

                        strrect.Offset(x, (int)(y + phr.GetMaxHeight() * 0.6));

                        if (this.debug)
                        { g.DrawRectangle(new Pen(Color.Red, 1), strrect); }

                        GraphicsPath path = new GraphicsPath();
                        path.AddString(w.str, fnt.FontFamily, (int)fnt.Style, strrect.Height,
                            new Rectangle(x, strrect.Y, this.width - x, (int)phr.GetMaxHeight()), sf);

                        Pen p = new Pen(w.cond.color[1], 5);
                        // RecENT の不具合？ - Tik (2007年04月10日 22時22分26秒)
                        p.LineJoin = System.Drawing.Drawing2D.LineJoin.Round;

                        g.DrawPath(p, path);
                        g.FillPath(new SolidBrush(w.cond.color[0]), path);

                        p.Color = w.cond.color[3];
                        g2.DrawPath(p, path);
                        g2.FillPath(new SolidBrush(w.cond.color[2]), path);

                        if (!String.IsNullOrEmpty(w.ruby))
                        {
                            Font rubyfont = w.cond.rubyfont.font;

                            StringFormat sfr = new StringFormat();
                            sfr.FormatFlags = StringFormatFlags.NoClip | StringFormatFlags.NoWrap;
                            sfr.Trimming = StringTrimming.None;

                            sfr.SetMeasurableCharacterRanges(new CharacterRange[] { new CharacterRange(0, w.ruby.Length) });
                            Region[] regiruby = g.MeasureCharacterRanges(w.ruby, rubyfont, new RectangleF(0, 0, this.width, this.height), sfr);
                            Rectangle recruby = Rectangle.Truncate(regiruby[0].GetBounds(g));

                            sfr.LineAlignment = w.cond.rubyfont.line_align;

                            if (recruby.Height <= 0) { goto skip_drawing; }
                            recruby.Width = (int)(recruby.Width * w.cond.rubyfont.adjust_width);

                            if (recruby.Width < strrect.Width) { sfr.Alignment = StringAlignment.Center; }
                            else { sfr.Alignment = StringAlignment.Near; }

                            GraphicsPath path_r = new GraphicsPath();
                            //strrect.Offset(0, -(fnt.Height / 2) - 3);
                            //path_r.AddString(w.ruby, fnt.FontFamily, (int)fnt.Style, fnt.Height / 2, strrect, sfr);
                            path_r.AddString(w.ruby, rubyfont.FontFamily, (int)rubyfont.Style,
                                recruby.Height, new Rectangle(x, y, strrect.Width, (int)(phr.GetMaxHeight() * 0.6)), sfr);

                            Pen pr = new Pen(w.cond.color[5], 2);
                            pr.LineJoin = System.Drawing.Drawing2D.LineJoin.Round;

                            g.DrawPath(pr, path_r);
                            g.FillPath(new SolidBrush(w.cond.color[4]), path_r);

                            pr.Color = w.cond.color[7];
                            g2.DrawPath(pr, path_r);
                            g2.FillPath(new SolidBrush(w.cond.color[6]), path_r);
                        }

                    }

                skip_drawing:

                    /*

                    int afterxtemp = 0;

                    if (w.time < time && time < phr.getWordEnd(w))
                    {
                        float allf = (float)(phr.getWordEnd(w).allFrames - w.time.allFrames);
                        int nowf = time.allFrames - w.time.allFrames;
                        afterxtemp = oldX + (int)(lsi * (nowf / allf));
                    }
                    else if (phr.getWordEnd(w) <= time)
                    {
                        afterxtemp = oldX + lsi;
                    }
                    if (afterxtemp > afterx) { afterx = afterxtemp; }
                    */
                    this.cache[phr.showline].drawed_X.Add(oldX);
                    x = oldX + lsi;
                } //word loop end

                this.cache[phr.showline].drawed_X.Add(x);

            }
            else
            {
                if (this.debug)
                    g_moto.DrawString("HIT! " + phr.showline.ToString(), new Font("MS UI Gothic", 20), Brushes.HotPink, new Point(10, 50 + 20 * phr.showline));
            }

            if (phr.words.Count + 1 != this.cache[phr.showline].drawed_X.Count)
            {
                throw new Exception("BUG: ワード数とキャッシュの対応がおかしいです");
            }

            int afterx;

            if (phr.getEndtime() < time)
            {
                afterx = this.width; //^^;
            }
            else
            {
                afterx = 0;
                for (int wcnt = 0; wcnt < phr.words.Count; wcnt++)
                {
                    if (phr.words[wcnt].time < time && time < phr.getWordEnd(phr.words[wcnt]))
                    {
                        float allf = (float)(phr.getWordEnd(phr.words[wcnt]).allFrames - phr.words[wcnt].time.allFrames);
                        int nowf = time.allFrames - phr.words[wcnt].time.allFrames;
                        afterx = this.cache[phr.showline].drawed_X[wcnt]
                            + (int)((this.cache[phr.showline].drawed_X[wcnt + 1] - this.cache[phr.showline].drawed_X[wcnt]) * (nowf / allf));
                        break;
                    }
                    else if (phr.getWordEnd(phr.words[wcnt]) <= time)
                    {
                        afterx = this.cache[phr.showline].drawed_X[wcnt + 1];
                    }
                }
            }

            if (afterx < 0) { afterx = 0; }
            if (afterx > this.width) { afterx = this.width; }
            
            //g_moto.DrawImage(bmp1, afterx, 0, new Rectangle(afterx, 0, this.width - afterx, this.height), GraphicsUnit.Pixel);
            //g_moto.DrawImageUnscaledAndClipped(bmp2, new Rectangle(0, 0, afterx, this.height));

            g_moto.DrawImage(this.cache[phr.showline].bmp_moto, afterx, 0, new Rectangle(afterx, 0, this.width - afterx, this.height), GraphicsUnit.Pixel);
            g_moto.DrawImageUnscaledAndClipped(this.cache[phr.showline].bmp_effe, new Rectangle(0, 0, afterx, this.height));

        }

    }

    public struct drawCache
    {
        public Phrase phr;
        public Bitmap bmp_moto;
        public Bitmap bmp_effe;
        public List<int> drawed_X;

        public drawCache(Phrase phrase)
        {
            this.phr = phrase;
            this.bmp_moto = null;
            this.bmp_effe = null;
            this.drawed_X = new List<int>();
        }
    }

}
